from enum import Enum

class CautionHourType(Enum):
    SUAVE = "suave"
    INTERMEDIATE = "intermediate"
    AGGRESSIVE = "aggressive"

    @staticmethod
    def get_num_value(type_state):
        if isinstance(type_state, str):
            tt = type_state.lower()
        elif isinstance(type_state, CautionHourType):
            tt = type_state.value
        else:
            print(f"type of {type_state} is {type(type_state)}")
            raise ValueError
        
        return VALUES_CONVERSION[tt]

VALUES_CONVERSION = {
            CautionHourType.SUAVE.value: 0.75,
            CautionHourType.INTERMEDIATE.value: 0.5,
            CautionHourType.AGGRESSIVE.value: 0.4
            }



from ..services.threshold.thresholdbase import ThresholdBase as t
from ..models.const import CURRENTS_THREEPHASE_1_32
import pytest

def test__start():
    ret = t._start(50, False)
    assert ret == 83.49

def test__start_caution_non_caution_late():
    ret = t._start(50, False)
    ret2 = t._start(50, True)
    assert ret == ret2

def test__start_caution_non_caution_early():
    ret = t._start(40, False)
    ret2 = t._start(40, True)
    assert ret > ret2

def test__stop():
    ret = t._stop(13, False)
    assert ret == 82.55

def test__stop_caution_non_caution_late():
    ret = t._stop(50, False)
    ret2 = t._stop(50, True)
    assert ret == ret2

def test__stop_caution_non_caution_early():
    ret = t._stop(40, False)
    ret2 = t._stop(40, True)
    assert ret > ret2
    
def test_allowed_current_base():
    ret = t.allowed_current(
        now_min=0,
        moving_avg=1,
        charger_enabled=False,
        charger_done=False,
        currents_dict=CURRENTS_THREEPHASE_1_32,
        total_energy=0,
        peak=1
    )
    assert ret == t.BASECURRENT

def test_allowed_current_1():
    ret = t.allowed_current(
        now_min=10,
        moving_avg=560,
        charger_enabled=True,
        charger_done=False,
        currents_dict=CURRENTS_THREEPHASE_1_32,
        total_energy=0.3,
        peak=10
    )
    assert ret == 16

def test_allowed_current_2():
    ret = t.allowed_current(
        now_min=50,
        moving_avg=560,
        charger_enabled=True,
        charger_done=False,
        currents_dict=CURRENTS_THREEPHASE_1_32,
        total_energy=0.3,
        peak=10
    )
    assert ret == 32


def test__start_quarterly():
    ret = t._start(50, False, True)
    assert ret == 60.62

def test__start_quarterly_caution():
    ret = t._start(50, True, True)
    assert ret == 52.13

def test__stop_quarterly():
    ret = t._start(22, False, True)
    assert ret == 65.29

def test__stop_quarterly_caution():
    ret = t._start(22, True, True)
    assert ret == 58.06


def test_allowed_current_negative_movingavg():
    ret = t.allowed_current(
        now_min=10,
        moving_avg=-560,
        charger_enabled=True,
        charger_done=False,
        currents_dict=CURRENTS_THREEPHASE_1_32,
        total_energy=0.3,
        peak=10
    )
    assert ret == 16

def test_allowed_current_negative_totalenergy():
    ret = t.allowed_current(
        now_min=10,
        moving_avg=560,
        charger_enabled=True,
        charger_done=False,
        currents_dict=CURRENTS_THREEPHASE_1_32,
        total_energy=-0.3,
        peak=10
    )
    assert ret == 16

def test_allowed_current_negative_totalenergy_and_movingavg():
    ret = t.allowed_current(
        now_min=10,
        moving_avg=-1560,
        charger_enabled=True,
        charger_done=False,
        currents_dict=CURRENTS_THREEPHASE_1_32,
        total_energy=-0.3,
        peak=10
    )
    assert ret == 20


from dataclasses import dataclass

@dataclass
class TieredPrice:
    lower_peak_limit: float
    value: float

    def __post_init__(self):
        assert self.value >= 0
        assert self.lower_peak_limit >= 0


from enum import Enum

class ChargeControllerStates(Enum):
    Idle = "Idle"
    Connected = "Connected"
    Start = "Start"
    Stop = "Stop"
    Done = "Done"
    Error = "Error, check logs."
    Charging = "Charging"
    Disabled = "Disabled"



from datetime import date, datetime, time
import logging

from ....models.locale.enums.querytype import QueryType

from .const import (
    QUERYTYPE_AVERAGEOFTHREEHOURS,
    QUERYTYPE_AVERAGEOFTHREEDAYS,
    QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19,
    QUERYTYPE_BASICMAX,
    QUERYTYPE_AVERAGEOFFIVEDAYS,
    QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR,
    QUERYTYPE_HIGHLOAD,
    QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22,
    QUERYTYPE_SOLLENTUNA
    )
from .queryservice import QueryService
from ....models.locale.peaks_model import PeaksModel
from ....models.locale.enums.sum_types import SumTypes
from ....models.locale.enums.time_periods import TimePeriods
from ....models.locale.sumcounter import SumCounter
from ....models.locale.queryproperties import QueryProperties

# from .querysets import QUERYSETS

_LOGGER = logging.getLogger(__name__)


class LocaleQuery:
    def __init__(
        self, 
        sum_type: SumTypes, 
        time_calc: TimePeriods, 
        cycle: TimePeriods, 
        sum_counter: SumCounter = None,
        #query_service: QueryService = None
        ) -> None:    
        self._peaks: PeaksModel = PeaksModel({})
        self._props = QueryProperties(
            sum_type, 
            time_calc, 
            cycle
            )
        self._sum_counter: SumCounter= sum_counter
        self._observed_peak_value: float = 0 
        self._charged_peak_value: float = 0

    def reset(self) -> None:
        self._peaks.reset()
        self._observed_peak_value = 0
        self._charged_peak_value = 0

    def set_query_service(self, service: QueryService) -> None:
        self._props.queryservice = service

    @property
    def peaks(self) -> PeaksModel:
        if self._peaks.is_dirty:
            self._sanitize_values()
        return self._peaks

    @property
    def sum_counter(self) -> SumCounter:
        if self._sum_counter is not None:
            return self._sum_counter
        return SumCounter()

    @property
    def charged_peak(self) -> float: 
        if self._peaks.is_dirty:
            self._sanitize_values()
        ret = self._charged_peak_value
        return round(ret,2)

    @charged_peak.setter
    def charged_peak(self, val):
        self._charged_peak_value = val

    @property
    def observed_peak(self) -> float: 
        if self._peaks.is_dirty:
            self._sanitize_values()
        ret = self.charged_peak if self._props.sumtype is SumTypes.Max else self._observed_peak_value
        return round(ret, 2)

    @observed_peak.setter
    def observed_peak(self, val):
        self._observed_peak_value = val

    def try_update(self, new_val, timestamp: datetime=datetime.now()):
        if self._props.queryservice.should_register_peak(dt=timestamp) is False:
            return
        _dt = (timestamp.day, timestamp.hour)
        if len(self.peaks.p) == 0:
            """first addition for this month"""
            self.peaks.add_kv_pair(_dt, new_val)
            self._peaks.set_month(timestamp.month)
        elif timestamp.month != self._peaks.m:
            """new month, reset"""
            self.reset_values(new_val, timestamp)
        else:
            self._set_update_for_groupby(new_val, _dt)
        if len(self.peaks.p) > self.sum_counter.counter:
                self.peaks.remove_min()
        self._update_peaks()

    def _set_update_for_groupby(self, new_val, dt):
        if self.sum_counter.groupby in [TimePeriods.Daily, TimePeriods.UnSet]:
            #todo: check this if it breaks the updatehour
            _datekeys = [k for k in self.peaks.p.keys() if dt[0] in k]
            if len(_datekeys):
                if new_val > self.peaks.p.get(_datekeys[0]):
                        self.peaks.pop_key(_datekeys[0])
                        self.peaks.add_kv_pair(dt, new_val)
            #todo: check this if it breaks the updatehour
            else:
                self.peaks.add_kv_pair(dt, new_val)
        elif self.sum_counter.groupby == TimePeriods.Hourly:
            if dt in self._peaks.p.keys():
                if new_val > self.peaks.p.get(dt):
                        self.peaks.add_kv_pair(dt, new_val)
            else:
                self.peaks.add_kv_pair(dt, new_val)

    def _update_peaks(self):
        if self._props.sumtype is SumTypes.Max:
            self.charged_peak = self._peaks.max_value
        elif self._props.sumtype is SumTypes.Avg:
            self.observed_peak = self._peaks.min_value
            self.charged_peak = self._peaks.value_avg

    def reset_values(self, new_val, dt = datetime.now()):
        self._peaks.clear()
        self.try_update(new_val, dt)

    def _sanitize_values(self):
        countX = lambda arr, x: len([a for a in arr if a[0] == x])
        if self.sum_counter.groupby == TimePeriods.Daily:
            duplicates = {}
            for k in self._peaks.p.keys():
                if countX(self._peaks.p.keys(), k[0]) > 1:
                    duplicates[k] = self._peaks.p[k]
            if len(duplicates):
                minkey = min(duplicates, key=duplicates.get)
                self._peaks.p.pop(minkey)
        while len(self._peaks.p) > self.sum_counter.counter:
            self._peaks.remove_min()
        self._peaks.is_dirty = False
        self._update_peaks()


QUERYTYPES = {
    QueryType.AverageOfThreeHours: LocaleQuery(sum_type=SumTypes.Avg, time_calc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sum_counter=SumCounter(counter=3, groupby=TimePeriods.Hourly)),
    QueryType.AverageOfThreeDays: LocaleQuery(sum_type=SumTypes.Avg, time_calc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sum_counter=SumCounter(counter=3, groupby=TimePeriods.Daily)),
    QueryType.Max: LocaleQuery(sum_type=SumTypes.Max, time_calc=TimePeriods.Hourly, cycle=TimePeriods.Monthly),
    QueryType.AverageOfFiveDays: LocaleQuery(sum_type=SumTypes.Avg, time_calc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sum_counter=SumCounter(counter=5, groupby=TimePeriods.Daily))
}



import pytest
from ..hub.hub import Hub
from ..hub.hub_options import HubOptions, Price
from ..services.locale.Locale import LOCALE_SE_GOTHENBURG
from ..models.hourselection.cautionhourtype import CautionHourType

OPTIONS_REGULAR = HubOptions(
    price=Price(
        price_aware=True,
        cautionhour_type=CautionHourType.SUAVE.value
    ),
    peaqev_lite=False,
    startpeaks={1:2, 2:2, 3:1.8, 4:1.5, 5:1.5, 6:1.5, 7:1.5, 8:1.5, 9:1.5, 10:1.8, 11:2, 12:2}
)
OPTIONS_REGULAR.locale = LOCALE_SE_GOTHENBURG

MOCK_CHARGER_OBJ = ""
MOCK_STATE_MACHINE = ""

def test_hub_init():
    h = Hub(options=OPTIONS_REGULAR,domain="test", state_machine=MOCK_STATE_MACHINE, chargerobj = MOCK_CHARGER_OBJ, is_test=True)
    assert h.domain == "test"
    assert h.hours.price_aware is True
    assert h.scheduler.active is False
    


from unittest import mock
import pytest
import time
from ..services.chargecontroller.chargecontrollerbase import ChargeControllerBase
from ..models.chargecontroller_states import ChargeControllerStates

_CHARGERSTATESMOCK = {
    ChargeControllerStates.Idle: ["idle"],
    ChargeControllerStates.Connected: ["connected"],
    ChargeControllerStates.Charging: ["charging"],
}
_NONHOURSMOCK = [5,6,7]

# def test_init_chargecontrollerbase_ok():
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK
#     )

#     print(c.latest_charger_start)
#     print(time.time())
#     time.sleep(0.1)
#     assert c.latest_charger_start < time.time()

# def test_init_chargecontrollerbase_init_fail_empty_statesdict():
#     with pytest.raises(AssertionError):
#         ChargeControllerBase(
#         charger_state_translation={},
#         non_hours=_NONHOURSMOCK
#     )

# def test_init_chargecontrollerbase_init_fail_empty_statesdict_type():
#     with pytest.raises(AssertionError):
#         ChargeControllerBase(
#         charger_state_translation={CHARGERSTATES.Idle: []},
#         non_hours=_NONHOURSMOCK
#     )

# def test_done_timeout_override():
#     mocktimeout = 60
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK,
#         timeout= mocktimeout
#     )

#     assert c.done_timeout == mocktimeout

# def test_is_timeout_true():
#     mocktimeout = 2
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK,
#         timeout= mocktimeout
#     )
#     time.sleep(mocktimeout)
    
#     assert c._is_timeout == True

# def test_is_timeout_false():
#     mocktimeout = 2
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK,
#         timeout= mocktimeout
#     )

#     assert c._is_timeout == False

def test_start_threshold_should_start():
    c = ChargeControllerBase(
        charger_state_translation=_CHARGERSTATESMOCK,
        non_hours=_NONHOURSMOCK
    )

    ret = c._below_start_threshold(
        predicted_energy=0.2,
        current_peak=1,
        threshold_start=0.6
    )
    assert ret is True

def test_start_threshold_should_not_start():
    c = ChargeControllerBase(
        charger_state_translation=_CHARGERSTATESMOCK,
        non_hours=_NONHOURSMOCK
    )
    
    ret = c._below_start_threshold(
        predicted_energy=0.8,
        current_peak=1,
        threshold_start=0.6
    )
    assert ret is False

def test_stop_threshold_should_stop():
    c = ChargeControllerBase(
            charger_state_translation=_CHARGERSTATESMOCK,
            non_hours=_NONHOURSMOCK
        )
    ret = c._above_stop_threshold(
        predicted_energy=0.8,
        current_peak=1,
        threshold_stop=0.6
    )
    assert ret is True

def test_stop_threshold_should_not_stop():
    c = ChargeControllerBase(
        charger_state_translation=_CHARGERSTATESMOCK,
        non_hours=_NONHOURSMOCK
    )
    
    ret = c._above_stop_threshold(
        predicted_energy=0.2,
        current_peak=1,
        threshold_stop=0.6
    )
    assert ret is False

def test_stop_threshold_should_not_stop():
    c = ChargeControllerBase(
        charger_state_translation=_CHARGERSTATESMOCK,
        non_hours=_NONHOURSMOCK
    )
    
    ret = c._above_stop_threshold(
        predicted_energy=2,
        current_peak=1.98,
        threshold_stop=60
    )
    assert ret is False

# def test_chargecontroller_is_charging_nonhour():
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK
#     )
    
#     ret = c.get_status(
#         charger_state="charging",
#         charger_enabled=True,
#         charger_done=False,
#         car_power_sensor=50,
#         total_energy_this_hour=0.5,
#         current_hour=6    
#     )
#     assert ret == CHARGERSTATES.Stop

# def test_chargecontroller_is_idle():
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK
#     )
    
#     ret = c.get_status(
#         charger_state="idle",
#         charger_enabled=False,
#         charger_done=False,
#         car_power_sensor=50,
#         total_energy_this_hour=0.5,
#         current_hour=10    
#     )
#     assert ret == CHARGERSTATES.Idle

# def test_chargecontroller_is_connected():
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK
#     )
    
#     ret = c.get_status(
#         charger_state="connected",
#         charger_enabled=False,
#         charger_done=False,
#         car_power_sensor=50,
#         total_energy_this_hour=0.5,
#         current_hour=10    
#     )
#     assert ret == CHARGERSTATES.Connected

# def test_chargecontroller_is_idle_nonhour():
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK
#     )
    
#     ret = c.get_status(
#         charger_state="idle",
#         charger_enabled=False,
#         charger_done=False,
#         car_power_sensor=50,
#         total_energy_this_hour=0.5,
#         current_hour=6    
#     )
#     assert ret == CHARGERSTATES.Idle

# def test_chargecontroller_is_connected_nonhour():
#     c = ChargeControllerBase(
#         charger_state_translation=_CHARGERSTATESMOCK,
#         non_hours=_NONHOURSMOCK
#     )
    
#     ret = c.get_status(
#         charger_state="connected",
#         charger_enabled=False,
#         charger_done=False,
#         car_power_sensor=50,
#         total_energy_this_hour=0.5,
#         current_hour=6    
#     )
#     assert ret == CHARGERSTATES.Connected


import logging
from datetime import datetime
from typing import Tuple
import statistics as stat
from ....models.hourselection.cautionhourtype import CautionHourType
from .hoursselection_helpers import HourSelectionHelpers
from .hoursselection_helpers import HourSelectionCalculations
from ....models.hourselection.hourobject import HourObject
from ....models.hourselection.hourselectionmodels import HourSelectionModel
from ....models.hourselection.hourtypelist import HourTypeList

_LOGGER = logging.getLogger(__name__)

ALLOWANCE_SCHEMA = {
    CautionHourType.get_num_value(CautionHourType.SUAVE): 1.15,
    CautionHourType.get_num_value(CautionHourType.INTERMEDIATE): 1.05,
    CautionHourType.get_num_value(CautionHourType.AGGRESSIVE): 1
}

class HourSelectionService:
    def __init__(self, parent, base_mock_hour: int = None):
        self.parent = parent
        self._mock_hour = base_mock_hour
        self._preserve_interim: bool = False
        self.calc = HourSelectionCalculations()
        self.helpers = HourSelectionHelpers()

    def update(self, caller: str = None) -> None:
        if self._preserve_interim and caller == "today":
            self.parent.model.hours.hours_today = self.parent.model.hours.hours_tomorrow
            self.parent.model.hours.hours_tomorrow = HourObject([], [], {})
            return

        hours, hours_tomorrow = self._interim_day_update(
            today=self._update_per_day(prices=self.parent.model.prices_today), 
            tomorrow=self._update_per_day(prices=self.parent.model.prices_tomorrow)
            )
        self.parent.model.hours.hours_today = self._add_remove_limited_hours(hours)
        self.parent.model.hours.hours_tomorrow = self._add_remove_limited_hours(hours_tomorrow)
        self.update_hour_lists()

    def _update_per_day(self, prices: list) -> HourObject:
        pricedict = {}
        if prices is not None and len(prices) > 1:
            pricedict = self.helpers.create_dict(prices)
            normalized_pricedict = self.helpers.create_dict(
                self.calc.normalize_prices(prices)
                )
            if stat.stdev(prices) > 0.05:
                ready_hours = self._determine_hours(
                    self.calc.rank_prices(
                        pricedict, 
                        normalized_pricedict,
                        self.parent.model.adjusted_average
                        ), 
                        prices
                        )
                ret= HourObject(
                    nh=ready_hours.nh, 
                    ch=ready_hours.ch, 
                    dyn_ch=ready_hours.dyn_ch, 
                    pricedict=pricedict
                    )
            else:
                ret= HourObject(nh=[], ch=[], dyn_ch={},pricedict=pricedict)
            ret.offset_dict=self._get_offset_dict(normalized_pricedict)
            return ret
        return HourObject([],[],{})

    def _get_offset_dict(self, normalized_hourdict: dict):
        ret = {}
        _prices = [p-min(normalized_hourdict.values()) for p in normalized_hourdict.values()]
        average_val = stat.mean(_prices)
        for i in range(0,24):
            try:
                ret[i] = round((_prices[i]/average_val) - 1,2)
            except:
                ret[i] = 1
        return ret

    def update_hour_lists(
        self, 
        listtype:HourTypeList = None,
        ) -> None:
        hour = self.set_hour()
        if listtype is not None:
            match listtype:
                case HourTypeList.NonHour:
                    self.parent.model.hours.update_non_hours(hour)
                case HourTypeList.CautionHour:
                    self.parent.model.hours.update_caution_hours(hour)
                case HourTypeList.DynCautionHour:
                    self.parent.model.hours.update_dynanmic_caution_hours(hour)   
                case _:
                    pass
        else:
            self.parent.model.hours.update_non_hours(hour)
            self.parent.model.hours.update_caution_hours(hour)
            self.parent.model.hours.update_dynanmic_caution_hours(hour)
            self.parent.model.hours.update_offset_dict()

    def _add_remove_limited_hours(self, hours: HourObject) -> HourObject:
        """Removes cheap hours and adds expensive hours set by user limitation"""
        if hours is None or all([len(hours.nh) == 0, len(hours.ch) == 0, len(hours.dyn_ch) == 0]):
            return HourObject([],[],{},offset_dict=hours.offset_dict,pricedict=hours.pricedict)
        hours.add_expensive_hours(self.parent.model.options.absolute_top_price)
        hours.remove_cheap_hours(self.parent.model.options.min_price)
        
        return hours

    def _determine_hours(self, price_list: dict, prices: list) -> HourObject:
        ret = HourObject([],[],{})
        for p in price_list:
            _permax = self.__set_charge_allowance(price_list[p]["permax"])
            if self.__should_be_cautionhour(price_list[p], prices):
                ret.ch.append(p)
                ret.dyn_ch[p] = round(_permax,2)
            else:
                ret.nh.append(p)
        return ret

    def __should_be_cautionhour(self, price_item, prices) -> bool:
        first = any([
                    float(price_item["permax"]) <= self.parent.model.options.cautionhour_type,
                    float(price_item["val"]) <= (sum(prices)/len(prices))
                ])
        second = (self.parent.model.current_peak > 0 and self.parent.model.current_peak*price_item["permax"] > 1) or self.parent.model.current_peak == 0
        return all([first, second])

    def __set_charge_allowance(self, price_input) -> float:
        return round(abs(price_input - 1), 2) * ALLOWANCE_SCHEMA[self.parent.model.options.cautionhour_type]
        
    def set_hour(self, testhour:int = None) -> int:
        return testhour if testhour is not None else self._mock_hour if self._mock_hour is not None else datetime.now().hour

    def _interim_day_update(self, today: HourObject, tomorrow: HourObject) -> Tuple[HourObject, HourObject]:
        """Updates the non- and caution-hours with an adjusted mean of 14h - 13h today-tomorrow to get a more sane nightly curve."""

        if len(self.parent.model.prices_tomorrow) < 23:
            return today, tomorrow 
        
        #hour = self._mock_hour if self._mock_hour is not None else 14
        hour =14
        negative_hour = (24 - hour)*-1

        pricelist = self.parent.model.prices_today[hour::]
        pricelist[len(pricelist):] = self.parent.model.prices_tomorrow[0:hour]
        new_hours = self._update_per_day(pricelist)
        
        today = self._update_interim_lists(range(hour,24), today, new_hours, hour)
        tomorrow = self._update_interim_lists(range(0,hour), tomorrow, new_hours, negative_hour)

        self._preserve_interim = True
        return today, tomorrow

    def _update_interim_lists(self, _range: range, old: HourObject, new: HourObject, index_devidation: int) -> HourObject:
        _new = HourSelectionHelpers._convert_collections(new, index_devidation)
        for i in _range:
            if i in _new.nh:
                if i not in old.nh:
                    old.nh.append(i)
                    old.ch = HourSelectionHelpers._try_remove(i, old.ch)
                    old.dyn_ch = HourSelectionHelpers._try_remove(i, old.dyn_ch)
            elif i in _new.ch:
                if i not in old.ch:
                    old.ch.append(i)
                    old.dyn_ch[i] = _new.dyn_ch[i]
                    old.nh = HourSelectionHelpers._try_remove(i, old.nh)
            else:
                old.nh = HourSelectionHelpers._try_remove(i, old.nh)
                old.ch = HourSelectionHelpers._try_remove(i, old.ch)
                old.dyn_ch = HourSelectionHelpers._try_remove(i, old.dyn_ch)
            old.nh.sort()
            old.ch.sort()
            old.dyn_ch = dict(sorted(old.dyn_ch.items()))
        
        for r in _new.offset_dict.keys():
            old.offset_dict[r] = _new.offset_dict[r]
            old.pricedict[r] = _new.pricedict[r]
        return old

    

    



from abc import abstractmethod
from ...util import _convert_quarterly_minutes
from datetime import datetime
import logging
from ...models.phases import Phases
from ...models.const import (
    CURRENTS_ONEPHASE_1_16, CURRENTS_THREEPHASE_1_16, CURRENTS_ONEPHASE_1_32, CURRENTS_THREEPHASE_1_32
)

_LOGGER = logging.getLogger(__name__)


class ThresholdBase:
    BASECURRENT = 6
    def __init__(self, hub):
        self._hub = hub
        self._phases = Phases.Unknown
        self._currents = {}

    @property
    def phases(self) -> str:
        return self._phases.name

    @property
    def currents(self) -> dict:
        return self._currents

    @property
    def stop(self) -> float:
        return ThresholdBase._stop(
            datetime.now().minute,
            str(datetime.now().hour) in self._hub.hours.caution_hours if self._hub.options.price.price_aware is False else False,
            self._hub.sensors.locale.data.is_quarterly(self._hub.sensors.locale.data)
        )

    @property
    def start(self) -> float:
        return ThresholdBase._start(
            datetime.now().minute,
            str(datetime.now().hour) in self._hub.hours.caution_hours if self._hub.options.price.price_aware is False else False,
            self._hub.sensors.locale.data.is_quarterly(self._hub.sensors.locale.data)
        )

    @property
    @abstractmethod
    def allowedcurrent(self) -> int:
        pass

    def _setcurrentdict(self) -> dict:
        """only allow amps if user has set this value high enough"""
        if self._hub.chargertype.max_amps != 16:
            _threephase = {k: v for (k, v) in CURRENTS_THREEPHASE_1_32.items() if v <= self._hub.chargertype.max_amps}
            _onephase = {k: v for (k, v) in CURRENTS_ONEPHASE_1_32.items() if v <= self._hub.chargertype.max_amps}
        else:
            _threephase = CURRENTS_THREEPHASE_1_16
            _onephase = CURRENTS_ONEPHASE_1_16

        try:
            divid = int(self._hub.sensors.carpowersensor.value)/int(self._hub.sensors.chargerobject_switch.current)
            if int(self._hub.sensors.carpowersensor.value) == 0:
                self._phases = Phases.Unknown
                ret = _threephase
            elif divid < 300:
                self._phases = Phases.OnePhase
                ret = _onephase
            else:
                self._phases = Phases.ThreePhase
                ret = _threephase
        except:
            _LOGGER.debug("Currents-dictionary: could not divide charger amps with charger power. Falling back to legacy-method.")
            if 0 < int(self._hub.sensors.carpowersensor.value) < 4000:
                self._phases = Phases.OnePhase
                ret = _onephase
            else:
                self._phases = Phases.ThreePhase
                ret = _threephase
        self._currents = ret
        return ret

    @staticmethod
    def _stop(
              now_min: int,
              is_caution_hour: bool,
              is_quarterly: bool=False
              ) -> float:
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        
        if is_caution_hour and minute < 45:
            ret = (((minute+pow(1.075, minute)) * 0.0032) + 0.7)
        else:
            ret = (((minute + pow(1.071, minute)) * 0.00165) + 0.8)
        return round(ret * 100, 2)

    @staticmethod
    def _start(
               now_min: int,
               is_caution_hour: bool,
               is_quarterly:bool=False
               ) -> float:
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        if is_caution_hour and minute < 45:
            ret = (((minute+pow(1.081, minute)) * 0.0049) + 0.4)
        else:
            ret = (((minute + pow(1.066, minute)) * 0.0045) + 0.5)
        return round(ret * 100, 2)
    
    @staticmethod
    def allowed_current(
            now_min: int,
            moving_avg: float,
            charger_enabled: bool,
            charger_done: bool,
            currents_dict: dict,
            total_energy: float,
            peak: float,
            is_quarterly:bool=False,
            power_canary_amp: int = -1
            ) -> int:
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        ret = ThresholdBase.BASECURRENT
        if not charger_enabled or charger_done or moving_avg == 0:
            return ret
        currents = currents_dict
        for key, value in currents.items():
            if ((((moving_avg + key) / 60) * (60 - minute) + total_energy * 1000) / 1000) < peak:
                ret = value
                break
        return min(ret, power_canary_amp) if power_canary_amp > -1 else ret



import statistics as stat
import logging
from ....models.hourselection.hourobject import HourObject

_LOGGER = logging.getLogger(__name__)


class HourSelectionHelpers:
    
    @staticmethod
    def create_dict(input: list):
        ret = {}
        for idx, val in enumerate(input):
            ret[idx] = val
        if 23 <= len(ret) <= 24:
            return ret
        elif len(ret) == 25:
            _LOGGER.debug(f"Looks like we are heading into DST. re-parsing hours")
            input.pop(2)
            return HourSelectionHelpers.create_dict(input)
        else:
            _LOGGER.exception(f"Could not create dictionary from pricelist: {input} with len {len(ret)}.")
            raise ValueError

    @staticmethod
    def convert_none_list(lst: any) -> list:
        ret = []
        if lst is None or not isinstance(lst, list):
            return ret
        try:
            for l in lst:
                if l is None:
                    return ret
            return lst
        except:
            return HourSelectionHelpers._make_array_from_empty(lst)

    

    @staticmethod
    def _try_parse(input:str, parsetype:type):
        try:
            ret = parsetype(input)
            return ret
        except:
            return False

    @staticmethod
    def _make_array_from_empty(input: str) -> list:
        array = input.split(",")
        list = [p for p in array if len(p)]
        ret = []
        if len(list) > 24:
            try:
                for l in list:
                    parsed_item = HourSelectionHelpers._try_parse(l, float)
                    if not parsed_item:
                        parsed_item = HourSelectionHelpers._try_parse(l, int)
                    assert isinstance(parsed_item, (float,int))
                    ret.append(parsed_item)
                return ret
            except:
                _LOGGER.warning("Unable to create empty list for prices.")
                pass
        return []

    

    @staticmethod
    def _convert_collections(new: HourObject, index_deviation: int) -> HourObject:
        """Converts the hourobject-collections to interim days, based on the index-deviation provided."""

        def _chop_list(lst: list):
            return [n+index_deviation for n in lst if 0 <= n+index_deviation < 24]
        def _chop_dict(dct: dict):
            return {key+index_deviation:value for (key,value) in dct.items() if 0 <= key+index_deviation < 24}
        ret = HourObject([], [], {})
        ret.nh = _chop_list(new.nh)
        ret.ch = _chop_list(new.ch)
        ret.dyn_ch = _chop_dict(new.dyn_ch)
        ret.offset_dict = _chop_dict(new.offset_dict)
        ret.pricedict = _chop_dict(new.pricedict)
        return ret

    @staticmethod
    def _try_remove(value, collection: list|dict):
        if isinstance(collection, dict):
            if value in collection.keys():
                collection.pop(value)
        elif isinstance(collection, list):
            if value in collection:
                collection.remove(value)
        return collection


class HourSelectionCalculations:
    
    def normalize_prices(self, prices:list) -> list:
        min_price = min(prices)
        c = 0
        if min_price <= 0:
            c = abs(min_price) + 0.01
        ret = []
        for p in prices:
            pp = p+c
            divider = min_price if min_price > 0 else c
            ret.append(round(pp-divider,3))
        return ret

    def rank_prices(self, hourdict: dict, normalized_hourdict: dict, adjusted_average:float = None) -> dict:
        """Rank the normalized pricelist to find out which are going to become non- or caution-hours"""
        
        CAUTIONHOUR_CUTOFF = 0.7 #cutoff from the average to allow prices below average to become cautionary.

        _adj_avg = stat.mean(hourdict.values()) if not isinstance(adjusted_average, (float,int)) else adjusted_average
        adj_average_norm = _adj_avg * (stat.mean(normalized_hourdict.values())/stat.mean(hourdict.values()))
        cautions = [h for h in normalized_hourdict if normalized_hourdict[h] > (adj_average_norm * CAUTIONHOUR_CUTOFF)]
        maxval = max(hourdict.values())
        ret = {}
        
        for c in cautions:
            ret[c] = {"val": hourdict[c], "permax": round(hourdict[c] / maxval,2)}

        return self._discard_excessive_hours(ret)

    def _discard_excessive_hours(self, hours: dict):
        """There should always be at least four regular hours before absolute_top_price kicks in."""
        while len(hours) >= 20:
            to_pop = dict(sorted(hours.items(), key=lambda item: item[1]['val']))    
            hours.pop(list(to_pop.keys())[0])
        return hours


from ...hub.hub_options import HubOptions
from .chargecontroller import ChargeController
from .chargecontroller_lite import ChargeControllerLite

class ChargeControllerFactory:
    @staticmethod
    def create(options: HubOptions):
        if options.peaqev_lite:
            return ChargeControllerLite(options)
        return ChargeController(options)